from __future__ import print_function

from terminaltables import AsciiTable
import colorama
from termcolor import colored
import six

from lemoncheesecake.helpers.text import wrap_text
from lemoncheesecake.helpers.time import humanize_duration
from lemoncheesecake.helpers import terminalsize
from lemoncheesecake.reporting import LogData, CheckData, UrlData, AttachmentData
from lemoncheesecake.testtree import flatten_tests
from lemoncheesecake.filter import filter_suites


def test_status_to_color(status):
    return {
        "passed": "green",
        "failed": "red",
        "disabled": "cyan",
        "in_progress": "cyan"
    }.get(status, "yellow")


def log_level_to_color(level):
    return {
        "debug": "cyan",
        "info": "cyan",
        "warn": "yellow",
        "error": "red"
    }.get(level, "cyan")


def outcome_to_color(outcome):
    return "green" if outcome else "red"


def _serialize_metadata(tags, properties, links, disabled):
    parts = []

    if disabled:
        parts.append("DISABLED")
    parts.extend(tags)
    parts.extend("%s:%s" % (key, properties[key]) for key in sorted(properties))
    parts.extend(link_name or link_url for link_url, link_name in links)

    return ", ".join(parts)


def serialize_metadata(obj, hide_disabled=False):
    return _serialize_metadata(
        obj.tags, obj.properties, obj.links,
        disabled=not hide_disabled and getattr(obj, "disabled", False)
    )


def serialize_hierarchy_metadata(obj, hide_disabled=False):
    return _serialize_metadata(
        obj.hierarchy_tags, obj.hierarchy_properties, obj.hierarchy_links,
        disabled=not hide_disabled and (hasattr(obj, "is_disabled") and obj.is_disabled())
    )


class Renderer(object):
    def __init__(self, max_width, explicit=False):
        self.max_width = max_width
        self.explicit = explicit
        # "20" is an approximation of the maximal overhead of table border, padding, and table first cell
        self._table_overhead = 20

    def wrap_description_col(self, description):
        return wrap_text(description, int((self.max_width - self._table_overhead) * 0.75))

    def wrap_details_col(self, details):
        return wrap_text(details, int((self.max_width - self._table_overhead) * 0.25))

    def render_check_outcome(self, outcome):
        if self.explicit:
            check_label = "CHECK %s" % ("OK" if outcome else "KO")
        else:
            check_label = "CHECK"

        return colored(check_label, color=outcome_to_color(outcome), attrs=["bold"])

    def render_steps(self, steps):
        rows = []
        for step in steps:
            rows.append([
                "",
                colored(
                    self.wrap_description_col(step.description),
                    color=outcome_to_color(step.is_successful()),
                    attrs=["bold"]
                ),
                colored(humanize_duration(step.duration, show_milliseconds=True), attrs=["bold"])
                    if step.duration is not None else "-"
            ])
            for entry in step.entries:
                if isinstance(entry, LogData):
                    rows.append([
                        colored(entry.level.upper(), color=log_level_to_color(entry.level), attrs=["bold"]),
                        self.wrap_description_col(entry.message)
                    ])
                if isinstance(entry, CheckData):
                    rows.append([
                        self.render_check_outcome(entry.outcome),
                        self.wrap_description_col(entry.description),
                        self.wrap_details_col(entry.details)
                    ])
                if isinstance(entry, UrlData):
                    rows.append([
                        colored("URL", color="cyan", attrs=["bold"]),
                        self.wrap_description_col("%s (%s)" % (entry.url, entry.description))
                    ])
                if isinstance(entry, AttachmentData):
                    rows.append([
                        colored("ATTACH", color="cyan", attrs=["bold"]),
                        self.wrap_description_col(entry.description),
                        entry.filename
                    ])

        table = AsciiTable(rows)
        table.inner_heading_row_border = False
        table.justify_columns[0] = "center"
        table.inner_row_border = True

        return table.table

    def render_result(self, description, short_description, status, steps):
        if status is None:
            status = "in_progress"

        if steps:
            details = self.render_steps(steps)
        else:
            details = "n/a"

        parts = [
            colored(
                "%s: %s" % (status.upper(), description) if self.explicit else description,
                color=test_status_to_color(status),
                attrs=["bold"]
            )
        ]
        if short_description:
            parts.append(colored("(%s)" % short_description, attrs=["bold"]))
        parts.append(details)

        return "\n".join(parts)

    def render_test(self, test):
        test_metadata = serialize_hierarchy_metadata(test, hide_disabled=True)
        if test_metadata:
            short_description = "%s - %s" % (test.path, test_metadata)
        else:
            short_description = test.path

        return self.render_result(test.description, short_description, test.status, test.steps)

    def render_tests(self, tests):
        for test in tests:
            yield self.render_test(test)
            yield ""

    def render_suite(self, suite):
        if suite.suite_setup:
            yield self.render_result(
                "- SUITE SETUP - %s" % suite.description, suite.path,
                suite.suite_setup.status, suite.suite_setup.steps
            )
            yield ""

        for test in suite.get_tests():
            yield self.render_test(test)
            yield ""

        if suite.suite_teardown:
            yield self.render_result(
                "- SUITE TEARDOWN - %s" % suite.description, suite.path,
                suite.suite_teardown.status, suite.suite_teardown.steps
            )
            yield ""

    def render_report(self, report):
        if report.test_session_setup:
            yield self.render_result(
                "- TEST SESSION SETUP -", None,
                report.test_session_setup.status, report.test_session_setup.steps
            )
            yield ""

        for suite in report.all_suites():
            for data in self.render_suite(suite):
                yield data

        if report.test_session_teardown:
            yield self.render_result(
                "- TEST SESSION TEARDOWN -", None,
                report.test_session_teardown.status, report.test_session_teardown.steps
            )
            yield ""


def _print_data(data_it):
    for data in data_it:
        print(data if six.PY3 else data.encode("utf8"))


def print_report(report, filtr=None, max_width=None, explicit=False):
    ###
    # Setup terminal
    ###
    colorama.init()
    if not max_width:
        max_width, _ = terminalsize.get_terminal_size()

    ###
    # Get a generator over data to be printed on the console
    ###
    renderer = Renderer(max_width=max_width, explicit=explicit)
    if not filtr or filtr.is_empty():
        if report.nb_tests == 0:
            print("No tests found in report")
            return
        data = renderer.render_report(report)
    else:
        suites = filter_suites(report.get_suites(), filtr)
        if not suites:
            print("The filter does not match any test in the report")
            return
        data = renderer.render_tests(flatten_tests(suites))

    ###
    # Do the actual job
    ###
    _print_data(data)



'''
Created on Mar 19, 2016

@author: nicolas
'''

import os
import os.path as p
from shutil import copy, copytree

from lemoncheesecake.helpers.resources import get_resource_path
from lemoncheesecake.reporting.backend import ReportingBackend, ReportingSession


class HtmlReportWriter(ReportingSession):
    def __init__(self, report_dir, offline_mode):
        self.report_dir = report_dir
        self.offline_mode = offline_mode

    def on_test_session_start(self, event):
        html_resource_dir = get_resource_path("html")
        report_resource_dir = p.join(self.report_dir, ".html")

        os.mkdir(report_resource_dir)
        copy(p.join(html_resource_dir, "report.js"), report_resource_dir)
        copy(p.join(html_resource_dir, "report.css"), report_resource_dir)

        if self.offline_mode:
            copy(p.join(html_resource_dir, "report_offline.html"), p.join(self.report_dir, "report.html"))
            copy(p.join(html_resource_dir, "jquery-1.12.3.min.js"), report_resource_dir)
            copytree(p.join(html_resource_dir, "bootstrap-3.3.6-dist"),
                     p.join(report_resource_dir, "bootstrap-3.3.6-dist"))
            copy(p.join(html_resource_dir, "bootstrap-slate.min.css"),
                 p.join(report_resource_dir, "bootstrap-3.3.6-dist", "css", "bootstrap.min.css"))
        else:
            copy(p.join(html_resource_dir, "report.html"), self.report_dir)


class HtmlBackend(ReportingBackend):
    name = "html"

    def __init__(self, offline_mode=False):
        self.offline_mode = offline_mode

    def create_reporting_session(self, report_dir, report, parallel, saving_strategy):
        return HtmlReportWriter(report_dir, self.offline_mode)



'''
Created on Jan 24, 2016

@author: nicolas
'''

import os.path
from contextlib import contextmanager
import shutil
import threading
import traceback

import six

from lemoncheesecake.exceptions import LemonCheesecakeInternalError
from lemoncheesecake.consts import ATTACHEMENT_DIR, \
    LOG_LEVEL_DEBUG, LOG_LEVEL_ERROR, LOG_LEVEL_INFO, LOG_LEVEL_WARN
from lemoncheesecake.reporting import *
from lemoncheesecake import events
from lemoncheesecake.exceptions import ProgrammingError

__all__ = "log_debug", "log_info", "log_warn", "log_warning", "log_error", "log_url", "log_check", \
    "set_step", "end_step", "detached_step", "Thread", \
    "prepare_attachment", "save_attachment_file", "save_attachment_content", \
    "prepare_image_attachment", "save_image_file", "save_image_content", \
    "add_report_info", "get_fixture"


_runtime = None  # type: _Runtime


def initialize_runtime(report_dir, report, scheduled_fixtures):
    global _runtime
    _runtime = _Runtime(report_dir, report, scheduled_fixtures)
    events.add_listener(_runtime)


class _Runtime(object):
    def __init__(self, report_dir, report, scheduled_fixtures):
        self.report_dir = report_dir
        self.report = report
        self.scheduled_fixtures = scheduled_fixtures
        self.attachments_dir = os.path.join(self.report_dir, ATTACHEMENT_DIR)
        self.attachment_count = 0
        self._attachment_lock = threading.Lock()
        self._failures = set()
        self._local = threading.local()
        self._local.location = None
        self._local.step = None

    def set_location(self, location):
        self._local.location = location

    @property
    def location(self):
        return self._local.location

    def mark_location_as_failed(self, location):
        self._failures.add(location)

    def is_successful(self, location):
        return location not in self._failures

    def is_everything_successful(self):
        return len(self._failures) == 0

    def set_step(self, description, detached=False):
        self._local.step = description
        events.fire(events.StepEvent(self._local.location, description, detached=detached))

    def end_step(self, step):
        events.fire(events.StepEndEvent(self._local.location, step))

    def _get_step(self, step):
        return step if step is not None else self._local.step

    def log(self, level, content, step=None):
        if level == LOG_LEVEL_ERROR:
            self.mark_location_as_failed(self.location)
        events.fire(events.LogEvent(self._local.location, self._get_step(step), level, content))

    def log_check(self, description, outcome, details, step=None):
        if outcome is False:
            self.mark_location_as_failed(self.location)
        events.fire(events.CheckEvent(self._local.location, self._get_step(step), description, outcome, details))

    def log_url(self, url, description, step=None):
        events.fire(events.LogUrlEvent(self._local.location, self._get_step(step), url, description))

    @contextmanager
    def prepare_attachment(self, filename, description, as_image=False, step=None):
        with self._attachment_lock:
            attachment_filename = "%04d_%s" % (self.attachment_count + 1, filename)
            self.attachment_count += 1
            if not os.path.exists(self.attachments_dir):
                os.mkdir(self.attachments_dir)

        yield os.path.join(self.attachments_dir, attachment_filename)

        events.fire(events.LogAttachmentEvent(
            self._local.location, self._get_step(step), "%s/%s" % (ATTACHEMENT_DIR, attachment_filename),
            filename, description, as_image
        ))

    def get_fixture(self, name):
        if not self.scheduled_fixtures.has_fixture(name):
            raise ProgrammingError("Fixture '%s' either does not exist or don't have a prerun_session scope" % name)

        try:
            return self.scheduled_fixtures.get_fixture_result(name)
        except AssertionError as excp:
            raise ProgrammingError(str(excp))


def get_runtime():
    # type: () -> _Runtime
    if not _runtime:
        raise LemonCheesecakeInternalError("Runtime is not initialized")
    return _runtime


def set_step(description, detached=False):
    """
    Set a new step.
    """
    get_runtime().set_step(description, detached=detached)


def end_step(step):
    """
    End a detached step.
    """
    get_runtime().end_step(step)


@contextmanager
def detached_step(description):
    set_step(description, detached=True)
    try:
        yield
    except Exception:
        # FIXME: use exception instead of last implicit stacktrace
        stacktrace = traceback.format_exc()
        if six.PY2:
            stacktrace = stacktrace.decode("utf-8", "replace")
        log_error("Caught unexpected exception while running test: " + stacktrace)
    end_step(description)


def log_debug(content, step=None):
    """
    Log a debug level message.
    """
    get_runtime().log(LOG_LEVEL_DEBUG, content, step=step)


def log_info(content, step=None):
    """
    Log a info level message.
    """
    get_runtime().log(LOG_LEVEL_INFO, content, step=step)


def log_warning(content, step=None):
    """
    Log a warning level message.
    """
    get_runtime().log(LOG_LEVEL_WARN, content, step=step)


log_warn = log_warning


def log_error(content, step=None):
    """
    Log an error level message.
    """
    get_runtime().log(LOG_LEVEL_ERROR, content, step=step)


def log_check(description, outcome, details=None, step=None):
    get_runtime().log_check(description, outcome, details, step=step)


def _prepare_attachment(filename, description=None, as_image=False, step=None):
    return get_runtime().prepare_attachment(filename, description or filename, as_image=as_image, step=step)


def prepare_attachment(filename, description=None, step=None):
    """
    Prepare a attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return _prepare_attachment(filename, description, step=step)


def prepare_image_attachment(filename, description=None, step=None):
    """
    Prepare an image attachment using a pseudo filename and an optional description.
    The function returns the real filename on disk that will be used by the caller
    to write the attachment content.
    """
    return _prepare_attachment(filename, description, as_image=True, step=step)


def _save_attachment_file(filename, description=None, as_image=False, step=None):
    with _prepare_attachment(os.path.basename(filename), description, as_image=as_image, step=step) as report_attachment_path:
        shutil.copy(filename, report_attachment_path)


def save_attachment_file(filename, description=None, step=None):
    """
    Save an attachment using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    return _save_attachment_file(filename, description, step=step)


def save_image_file(filename, description=None, step=None):
    """
    Save an image using an existing file (identified by filename) and an optional
    description. The given file will be copied.
    """
    return _save_attachment_file(filename, description, as_image=True, step=step)


def _save_attachment_content(content, filename, description=None, as_image=False, binary_mode=False, step=None):
    with _prepare_attachment(filename, description, as_image=as_image, step=step) as report_attachment_path:
        with open(report_attachment_path, "wb") as fh:
            fh.write(content if binary_mode else content.encode("utf-8"))


def save_attachment_content(content, filename, description=None, binary_mode=False, step=None):
    """
    Save a given content as attachment using pseudo filename and optional description.
    """
    return _save_attachment_content(content, filename, description, binary_mode=binary_mode, as_image=False, step=step)


def save_image_content(content, filename, description=None, step=None):
    """
    Save a given image content as attachment using pseudo filename and optional description.
    """
    return _save_attachment_content(content, filename, description, binary_mode=True, as_image=True, step=step)


def log_url(url, description=None, step=None):
    """
    Log an URL.
    """
    get_runtime().log_url(url, description or url, step=step)


def get_fixture(name):
    """
    Return the corresponding fixture value. Only fixtures whose scope is session_prerun can be retrieved.
    """
    return get_runtime().get_fixture(name)


def add_report_info(name, value):
    report = get_runtime().report
    report.add_info(name, value)


def set_runtime_location(location):
    get_runtime().set_location(location)


def mark_location_as_failed(location):
    get_runtime().mark_location_as_failed(location)


def is_location_successful(location):
    return get_runtime().is_successful(location)


def is_everything_successful():
    return get_runtime().is_everything_successful()


class Thread(threading.Thread):
    def __init__(self, *args, **kwargs):
        super(Thread, self).__init__(*args, **kwargs)
        self.location = get_runtime().location

    def run(self):
        set_runtime_location(self.location)
        return super(Thread, self).run()



import sys

from lemoncheesecake.cli.command import Command
from lemoncheesecake.cli.utils import auto_detect_reporting_backends, add_report_path_cli_arg, get_report_path
from lemoncheesecake.reporting import load_report
from lemoncheesecake.reporting.backends.console import print_report_as_test_run
from lemoncheesecake.reporting.console import print_report
from lemoncheesecake.filter import add_report_filter_cli_args, make_report_filter


class ReportCommand(Command):
    def get_name(self):
        return "report"

    def get_description(self):
        return "Display a report"

    def add_cli_args(self, cli_parser):
        group = cli_parser.add_argument_group("Display report")
        add_report_path_cli_arg(group)
        group.add_argument(
            "--short", "-s", action="store_true", required=False,
            help="Display report as lcc run display test results"
        )
        group.add_argument(
            "--explicit", "-e", action="store_true", required=False,
            help="Make all indicators 'explicit' (i.e not only relying on a color-code), "
                 "will be enforced is stdout is redirected"
        )
        group.add_argument(
            "--max-width", "-w", type=int, required=False,
            help="Set a max width for tables (default is current terminal width)"
        )

        add_report_filter_cli_args(cli_parser)

    def run_cmd(self, cli_args):
        report_path = get_report_path(cli_args)
        report = load_report(report_path, auto_detect_reporting_backends())
        filtr = make_report_filter(cli_args)

        if cli_args.short:
            print_report_as_test_run(report, filtr)
        else:
            print_report(
                report, filtr=filtr, max_width=cli_args.max_width,
                explicit=cli_args.explicit or not sys.stdout.isatty()
            )

        return 0
